/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 OW2 consortium (www.ow2.org)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XMLUtil.java 98 2006-02-24 16:18:48Z alouis $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.util;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

/**
 * <p>
 * Document builder factories.
 * </p>
 * <p>
 * To avoid several instance of the same DocumentBuilder (with same options) in
 * the same Thread, each DocumentBulder must be created using these factories.
 * </p>
 * 
 * @author Christophe DENEUX (Capgemini Sud)
 * 
 */
public final class DocumentBuilders {
    
    /**
     * <p>
     * DocumentBuilder as thread local.
     * </p>
     * <p>
     * The document builder is the one provided by the JVM, even if an other DOM
     * implementation is in the classpath (ex: Xerces). The JVM implementation
     * is needed because the implementation of the DocumentFragment must be
     * desrializable on other Petals node, that have not a specific
     * implementation in its classpath.
     * </p>
     */
    private final static ThreadLocal<DocumentBuilder> jvmDocumentBuilderThreadLocal = new ThreadLocal<DocumentBuilder>() {

        @Override
        protected DocumentBuilder initialValue() {
            final ClassLoader currentClassLoader = Thread.currentThread().getContextClassLoader();
            try {
                final ClassLoader systemClassLoader = ClassLoader.getSystemClassLoader();
                Thread.currentThread().setContextClassLoader(systemClassLoader);
                
                final DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
                documentBuilderFactory.setNamespaceAware(true);
                final DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
                return documentBuilder;
            } catch (ParserConfigurationException e) {
                throw new RuntimeException("Failed to create DocumentBuilder", e);
            } finally {
                Thread.currentThread().setContextClassLoader(currentClassLoader);
            }
        }
    };
    
    /**
     * <p>
     * DocumentBuilder as thread local, supporting XML namespaces.
     * </p>
     * <p>
     * The document builder is the one provided by the classpath.
     * </p>
     */
    private final static ThreadLocal<DocumentBuilder> defaultDocumentBuilderThreadLocal = new ThreadLocal<DocumentBuilder>() {

        @Override
        protected DocumentBuilder initialValue() {
            try {
                final DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory
                        .newInstance();
                documentBuilderFactory.setNamespaceAware(true);
                final DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
                return documentBuilder;
            } catch (ParserConfigurationException e) {
                throw new RuntimeException("Failed to create DocumentBuilder", e);
            }
        }
    };

    /**
     * <p>
     * DocumentBuilder as thread local.
     * </p>
     * <p>
     * The document builder is the one provided by the JVM, even if an other DOM
     * implementation is in the classpath (ex: Xerces). The JVM implementation
     * is needed because the implementation of the DocumentFragment must be
     * desrializable on other Petals node, that have not a specific
     * implementation in its classpath.
     * </p>
     */
    public final static DocumentBuilder getJvmDocumentBuilder() {
        return DocumentBuilders.jvmDocumentBuilderThreadLocal.get();
    }

     /**
     * <p>
     * DocumentBuilder as thread local, supporting XML namespaces.
     * </p>
     * <p>
     * The document builder is the one provided by the classpath.
     * </p>
     */
     public final static DocumentBuilder getDefaultDocumentBuilder() {
        return DocumentBuilders.defaultDocumentBuilderThreadLocal.get();
    }

    
}
